"use strict";
// Copyright (c) Frédéric Maquin <fred@ephread.com>
// Licensed under the MIT License.
// See LICENSE in the project root for license information.
Object.defineProperty(exports, "__esModule", { value: true });
const ChildProcess = require("child_process");
const Fs = require("fs-extra");
const Os = require("os");
const Path = require("path");
const Uuid = require("uuid/v4");
const types_1 = require("./types");
const utils_1 = require("./utils");
/* Constants */
/******************************************************************************/
const INK_EXTENSIONS = ["ink", "ink2"];
/**
 * The global settings, used when the `workspace/configuration` request is not supported
 * by the client.
 */
exports.DEFAULT_SETTINGS = {
    inklecateExecutablePath: utils_1.defaultInklecatePath(utils_1.determinePlatform()),
    mainStoryPath: "main.ink",
    runThroughMono: utils_1.isRunThroughMono()
};
/* Helpers */
/******************************************************************************/
/**
 * Prepare the workspace for compilation. Create a temporary directory and
 * copies all of the project ink files into it, while preserving the subdirectory
 * hierarchy. The resulting directory should be kept safely for the remaining
 * of the session.
 *
 * @param workspace the workspace containing all the ink files.
 * @param connection the client connection.
 * @returns the temporary directory containing copies of the workspace's ink files.
 */
function prepareTempDirectoryForCompilation(workspaceFolder, logger, callback) {
    logger.log(`Creating temporary compile directory for: ${workspaceFolder.name}`);
    const tempDirectory = Path.join(Os.tmpdir(), Uuid());
    const workspacePath = utils_1.getPathFromUri(workspaceFolder.uri);
    // Make the temporary directory and copy the ink files in it.
    Fs.mkdir(tempDirectory, mkDirError => {
        if (mkDirError) {
            logger.log(mkDirError.message);
            callback(undefined);
        }
        else {
            Fs.copy(workspacePath, tempDirectory, {
                filter: (src, dest) => {
                    try {
                        const isDir = Fs.lstatSync(src).isDirectory();
                        const isInk = INK_EXTENSIONS.indexOf(src.split(".").pop() || "") > -1;
                        return isDir || isInk;
                    }
                    catch (error) {
                        logger.log(error.message);
                        return false;
                    }
                }
            }, copyError => {
                if (copyError) {
                    logger.log(copyError.message);
                    callback(undefined);
                }
                else {
                    callback(tempDirectory);
                }
            });
        }
    });
}
exports.prepareTempDirectoryForCompilation = prepareTempDirectoryForCompilation;
/**
 * Replace the original ink file found in `workspace.temporaryCompilationDirectory`,
 * with the update content of `document`.
 *
 * @param document document holding the new content.
 * @param workspace the workspace to which the document belongs.
 * @param callback called once the file has been updated.
 */
function updateFile(document, workspace, callback) {
    if (!workspace.folder || !workspace.temporaryCompilationDirectory) {
        return;
    }
    const workspacePath = utils_1.getPathFromUri(workspace.folder.uri);
    const documentContent = document.getText();
    const path = utils_1.getPathFromUri(document.uri);
    const regex = new RegExp(`^${workspacePath}`);
    const relativePath = path.replace(regex, "");
    const fullPath = Path.join(workspace.temporaryCompilationDirectory, relativePath);
    Fs.writeFile(fullPath, documentContent, callback);
}
exports.updateFile = updateFile;
/**
 * Compile the project and report any errors through `completion`.
 *
 * @param settings the configuration settings to use.
 * @param inkWorkspace the workspace to compile.
 * @param connection the client connection.
 * @param completion the callback to call upon completion.
 */
function compileProject(settings, inkWorkspace, logger, completion) {
    const tempDir = inkWorkspace.temporaryCompilationDirectory;
    if (!tempDir) {
        logger.log(`Temporary directory for ${inkWorkspace.folder.name} is \`undefined\`, ignoring…`);
        return;
    }
    const mergedSettings = utils_1.mergeSettings(settings, exports.DEFAULT_SETTINGS);
    const mainStoryTempPath = Path.join(tempDir, mergedSettings.mainStoryPath);
    // Though checking these beforehand is subjected to race conditions, it's useful as it'll
    // give the user a hint about what may be going wrong.
    Fs.access(mergedSettings.inklecateExecutablePath, Fs.constants.X_OK)
        .catch(error => {
        const inklecateErrorMessage = `'inklecatePath' (${mergedSettings.inklecateExecutablePath}) is not executable.`;
        logger.log(error.message);
        logger.showErrorMessage(inklecateErrorMessage);
    })
        .then(() => {
        return Fs.access(mainStoryTempPath, Fs.constants.R_OK);
    })
        .catch(error => {
        const storyErrorMessage = `'mainStoryPath' (${mergedSettings.mainStoryPath}) is not readable.`;
        logger.log(error.message);
        logger.showErrorMessage(storyErrorMessage);
    })
        .then(() => {
        spawnInklecate(mergedSettings, mainStoryTempPath, inkWorkspace, logger, completion);
    });
}
exports.compileProject = compileProject;
/**
 * Spawn a child process of `inklecatePath` to cimpile the project.
 *
 * @param inklecatePath the path to the inklecate executable.
 * @param mainStoryPath the path to the main ink file (in the client workspace).
 * @param mainStoryTempPath the path to the main ink file (in the temporary compile directory).
 * @param inkWorkspace the workspace to compile.
 * @param connection the client connection.
 * @param completion the callback to call upon completion.
 */
function spawnInklecate(settings, mainStoryTempPath, inkWorkspace, logger, completion) {
    // We'll let the process crash if inklecate could not be spawned.
    const command = settings.runThroughMono ? "mono" : settings.inklecateExecutablePath;
    const args = settings.runThroughMono
        ? [settings.inklecateExecutablePath, mainStoryTempPath]
        : [mainStoryTempPath];
    const inklecateProcess = ChildProcess.spawn(command, args, {
        cwd: Path.dirname(settings.inklecateExecutablePath),
        env: {
            MONO_BUNDLED_OPTIONS: "--debug"
        }
    });
    inklecateProcess.stderr.setEncoding("utf8");
    inklecateProcess.stderr.on("data", text => {
        if (typeof text === "string") {
            // Strip Byte order mark
            text = text.replace(/^\uFEFF/, "");
            if (text.length > 0) {
                const inklecateMessage = "Inklecate returned an error, see the log for more details.";
                logger.log(text);
                logger.showErrorMessage(inklecateMessage);
            }
        }
    });
    inklecateProcess.stdout.setEncoding("utf8");
    inklecateProcess.stdout.on("data", text => {
        if (typeof text === "string") {
            const errors = parseInklecateOutput(text, Path.dirname(settings.mainStoryPath), inkWorkspace);
            completion(errors);
        }
    });
}
/**
 * Parse the output of Inklecate and returns any errors or warnings found.
 *
 * @param text inklecate's output
 * @param mainStoryPathPrefix the prefix path for the real files locations (i. e. not the
 *                            temporary directory).
 * @param workspace the workspace for which the compilation took place.
 */
function parseInklecateOutput(text, mainStoryPathPrefix, workspace) {
    // Strip Byte order mark
    text = text.replace(/^\uFEFF/, "");
    if (text.length === 0) {
        return [];
    }
    const lines = text.split("\n");
    const inkErrors = [];
    for (const line of lines) {
        const trimmedLine = line.trim();
        const errorMatches = trimmedLine.match(/^(ERROR|WARNING|RUNTIME ERROR|RUNTIME WARNING|TODO): ('([^']+)' )?line (\d+): (.+)/);
        if (errorMatches) {
            const errorType = types_1.InkErrorType.parse(errorMatches[1]);
            const path = "file://" +
                Path.join(utils_1.getPathFromUri(workspace.folder.uri), mainStoryPathPrefix, errorMatches[3]);
            if (errorType) {
                inkErrors.push({
                    filePath: path,
                    lineNumber: parseInt(errorMatches[4]),
                    message: errorMatches[5],
                    type: errorType
                });
            }
        }
        else if (line.length > 0) {
            break;
        }
    }
    return inkErrors;
}
//# sourceMappingURL=compile.js.map